#include "pdp2.h"

void freqanalysis();

/************************************************************************/
SCALAR revers(int num, int n)
{
  SCALAR power, rev;
  int inum, iquot, irem;
  
  rev = 0.;
  inum = num;
  power = 1.;
  
  do {
    iquot = inum/n;
    irem = inum - n*iquot;
    power /= n;
    rev += irem*power;
    inum = iquot;
  } while (inum > 0);
  
  return (rev);
}

/************************************************************************/
/* frand() returns values 0 through 1.                                  */
/* From "Random number generators: good ones are hard to find", S. Park */
/* and K. Miller, Communications of ACM, October 1988, pp 1192-1201.    */
/* This is from page 1195, and is to work on any system for which       */
/* maxint is 2**31-1 or larger. Due earlier to Schrage, as cited by P&M.*/
/*                                                                      */
/* Note: OK values for iseed are 1 through 2**31-2. Give it 0 or 2*31-1 */
/* and it will return the same values thereafter!                       */
/*                                                                      */
/* C version 6/91, Bruce Langdon.                                       */
/*                                                                      */
/* Algorithm replaces seed by mod(a*seed,m). First represent            */
/* seed = q*hi + lo.  Then                                              */
/* a*seed = a*q*hi + lo = (m - r)*hi + a*lo = (a*lo - r*hi) + m*hi,     */
/* and new seed = a*lo - r*hi unless negative; if so, then add m.       */

SCALAR frand()
{
  long a = 16807, m = 2147483647, q = 127773, r = 2836;
  long hi, lo;
  SCALAR fnumb;
  /* static long seed=31207321; */
  
  hi = seed/q;
  lo = seed - q*hi;
  seed = a*lo - r*hi;
  /* "seed" will always be a legal integer of 32 bits (including sign). */
  if(seed <= 0) seed = seed + m;
  fnumb = seed/2147483646.0;

  return(fnumb);
}

/***********************************************************************/
/* time history accumulator; calculates and stores all history values, */
/* and performs combing on history values when low on memory           */

void history()
{
  register int i, j, k, isp;
  static int count=1, nrftcount=0;
  extern SCALAR lhs_source(), rhs_source();
  
  /**************************************/
  
  if(nfft) {
    if (thist_hi >= nfft) {
      freqanalysis();
      thist_hi=0;
    }
    phi_mid_hist[thist_hi] = phi[ncx/2][ncy/2];
    
    /*djc new */
    if(ybc_flag == PERIODIC) {
      phi_mid_k_t[0][2*thist_hi]       = phik[ncx/2][0];
      phi_mid_k_t[0][2*thist_hi + 1]   = 0.0;
      phi_mid_k_t[khi][2*thist_hi]     = phik[ncx/2][1];
      phi_mid_k_t[khi][2*thist_hi + 1] = 0.0;
      
      for(k=1;k<khi;k++) {
	phi_mid_k_t[k][2*thist_hi]     = phik[ncx/2][2*k];
	phi_mid_k_t[k][2*thist_hi + 1] = phik[ncx/2][2*k+1];
      }
    }
    
    if(lhs_flag) {
      phi_lhs_hist[thist_hi] = lhs_phi0;
      cur_lhs_hist[thist_hi] = lhs_exti;
      pow_lhs_hist[thist_hi] = lhs_exti*lhs_phi0;
    }
    if(rhs_flag) {
      phi_rhs_hist[thist_hi] = rhs_phi0;
      cur_rhs_hist[thist_hi] = rhs_exti;
      pow_rhs_hist[thist_hi] = rhs_exti*rhs_phi0;
    }
    Local_t_array[thist_hi] = t;
    thist_hi++;
    
    /********************************************/
    /****  AVERAGES OVER NFFT TIME STEPS   ******/
    
    if(nrftcount == nfft) {
      for(isp=0; isp<nsp; isp++)
	for (i=0; i<= ncx; i++)
	  for (j=0; j<= ncy; j++) {
	    sp_n_ave_show[isp][i][j] = sp_n_ave[isp][i][j]/nfft;
	    sp_n_ave[isp][i][j]= sp_n[isp][i][j];
	  }
      for (i=0; i<= ncx; i++)
	for (j=0; j<= ncy; j++) {
	  phi_ave_show[i][j]= phi_ave[i][j]/nfft;
	  phi_ave[i][j]= phi[i][j];
	}
      nrftcount =0;
    }
    else {
      for (i=0; i<= ncx; i++) {
	for (j=0; j<= ncy; j++) {
	  for(isp=0; isp<nsp; isp++)
	    sp_n_ave[isp][i][j]+= sp_n[isp][i][j];
	  
	  phi_ave[i][j] += phi[i][j];
	}
      }
      nrftcount++;
    }
  }
  
  if(ybc_flag == PERIODIC) {
    /*new djc*/
    for(i=0; i<=ncx;i++) {
      show_phik[i][0]=fabs(phik[i][0]);
      show_phik[i][khi]=fabs(phik[i][1]);
      show_rhok[i][0]=fabs(rhok[i][0]);
      show_rhok[i][khi]=fabs(rhok[i][1]);
      for(k=1;k<khi; k++) {
	show_phik[i][k]=sqrt(phik[i][2*k]*phik[i][2*k] + phik[i][2*k+1]*phik[i][2*k+1]);
	show_rhok[i][k]=sqrt(rhok[i][2*k]*rhok[i][2*k] + rhok[i][2*k+1]*rhok[i][2*k+1]);
      }
    }
  }
  
  /**************************************/
  
  if (--count) return;		/* only accum every interval steps */
  if (hist_hi >= HISTMAX) {	/* comb time histories */
    for (i=1, k=4; i<HISTMAX/4; i++, k+=4) {
      t_array[i] = t_array[k];
      com_phi_mid_hist[i] = com_phi_mid_hist[k];
    }
    if(lhs_flag) {
      for (i=1, k=4; i<HISTMAX/4; i++, k+=4) {
	com_phc_lhs_hist[i] = com_phc_lhs_hist[k];
	com_phs_lhs_hist[i] = com_phs_lhs_hist[k];
	com_phi_lhs_hist[i] = com_phi_lhs_hist[k];
	com_cur_lhs_hist[i] = com_cur_lhs_hist[k];
	com_pow_lhs_hist[i] = com_pow_lhs_hist[k];
      }
    }
    if(rhs_flag) {
      for (i=1, k=4; i<HISTMAX/4; i++, k+=4) {
	com_phc_rhs_hist[i] = com_phc_rhs_hist[k];
	com_phs_rhs_hist[i] = com_phs_rhs_hist[k];
	com_phi_rhs_hist[i] = com_phi_rhs_hist[k];
	com_cur_rhs_hist[i] = com_cur_rhs_hist[k];
	com_pow_rhs_hist[i] = com_pow_rhs_hist[k];
      }
    }
    for (isp=0; isp<nsp; isp++) {
      for (i=1, k=4; i<HISTMAX/4; i++, k+=4) {
	if(nstrcmax>0) { 
	  for(j=0; j<nstrctype; j++)
	    stc_np_hist[j][isp][i] = stc_np_hist[j][isp][k];
	}
	np_hist[isp][i] = np_hist[isp][k];
	kes_hist[isp][i]= kes_hist[isp][k];
      }
    }
    hist_hi = i;
    interval *= 4;  
  }
  
  t_array[hist_hi]  = t;            /* accumulate histories */
  com_phi_mid_hist[hist_hi] = phi[ncx/2][ncy/2];
  if(lhs_flag) {
    com_phc_lhs_hist[hist_hi] = lhs_extq/lhs_extc;
    com_phs_lhs_hist[hist_hi] = lhs_phi0 -phi[ncx/2][ncy/2];
    com_phi_lhs_hist[hist_hi] = lhs_phi0;
    com_cur_lhs_hist[hist_hi] = lhs_exti;
    com_pow_lhs_hist[hist_hi] = lhs_exti*lhs_phi0;
  }
  if(rhs_flag) {
    com_phc_rhs_hist[hist_hi] = rhs_extq/rhs_extc;
    com_phs_rhs_hist[hist_hi] = rhs_phi0 -phi[ncx/2][ncy/2];
    com_phi_rhs_hist[hist_hi] = rhs_phi0;
    com_cur_rhs_hist[hist_hi] = rhs_exti;
    com_pow_rhs_hist[hist_hi] = rhs_exti*rhs_phi0;
  }
  for (isp=0; isp<nsp; isp++) {
    np_hist[isp][hist_hi] = np[isp];
    if(nstrcmax>0) { 
      for(j=0; j<nstrctype; j++)
	stc_np_hist[j][isp][hist_hi] = stc_np[j][isp];
    }
  }
  hist_hi++;
  count = interval;
}

/***************************************************************/

void freqanalysis()
{
  register int i, j, k;
  static int init_freq_flag=1;
  static SCALAR *temp1, *temp2;
  
  if(init_freq_flag) {
    if(!(temp1= (SCALAR *)malloc((nfft/2)*sizeof(SCALAR)))
       || !(temp2= (SCALAR *)malloc((nfft/2)*sizeof(SCALAR))))
      puts("Null ptr in freqanalysis()");
    init_freq_flag=0;
  }
  /********** First do the MID potential   ************/
  for(i=0; i< nfft; i++)
    phi_mid_fft[i]= phi_mid_hist[i];
  
  realft(phi_mid_fft-1, freq_hi, 1);
  
  /**** Computing mag and phase of the voltage signal ***/
 
  temp1[0]= fabs(phi_mid_fft[0]);
  temp2[0]= (phi_mid_fft[0] > 0.0) ? 0.0 : 180.0;
  for(i=1, j=2; i< freq_hi; i++, j+=2) {
    temp1[i]= 2*sqrt(phi_mid_fft[j]*phi_mid_fft[j]+ phi_mid_fft[j+1]*phi_mid_fft[j+1]);
    if(fabs(phi_mid_fft[j+1]) < 1e-30 && fabs(phi_mid_fft[j]) < 1e-30)
      temp2[i]= 0.0;
    else
      temp2[i]= (180./M_PI)*atan2(phi_mid_fft[j], phi_mid_fft[j+1]);
  }
  for(i=0; i< freq_hi; i++) {
    phi_mid_fft[i]= temp1[i];
    phi_mid_fft[freq_hi+i]= temp2[i];
  }
  /********** Then do the LHS stuff     ************/
  if(lhs_flag) {
    for(i=0; i< nfft; i++) {
      cur_lhs_fft[i]= cur_lhs_hist[i];
      pow_lhs_fft[i]= pow_lhs_hist[i];
      phi_lhs_fft[i]= phi_lhs_hist[i];
    }
    realft(phi_lhs_fft-1, freq_hi, 1);
    realft(pow_lhs_fft-1, freq_hi, 1);
    realft(cur_lhs_fft-1, freq_hi, 1);
    
    /**** Computing mag and phase of the current signal ***/
    
    temp1[0]= fabs(cur_lhs_fft[0]);
    temp2[0]= (cur_lhs_fft[0] > 0.0) ? 0.0 : 180.0;
    for(i=1, j=2; i< freq_hi; i++, j+=2) {
      temp1[i]= 2*sqrt(cur_lhs_fft[j]*cur_lhs_fft[j]+ cur_lhs_fft[j+1]*cur_lhs_fft[j+1]);
      if(fabs(cur_lhs_fft[j+1]) < 1e-30 && fabs(cur_lhs_fft[j]) < 1e-30)
	temp2[i]= 0.0;
      else
	temp2[i]= (180./M_PI)*atan2(cur_lhs_fft[j], cur_lhs_fft[j+1]);
    }
    for(i=0; i< freq_hi; i++) {
      cur_lhs_fft[i]= temp1[i];
      cur_lhs_fft[freq_hi+i]= temp2[i];
    }
    
    /**** Computing mag and phase of the power signal *****/
    
    temp1[0]= fabs(pow_lhs_fft[0]);
    temp2[0]= (pow_lhs_fft[0] > 0.0) ? 0.0 : 180.0;
    for(i=1, j=2; i< freq_hi; i++, j+=2) {
      temp1[i]= 2*sqrt(pow_lhs_fft[j]*pow_lhs_fft[j]+ pow_lhs_fft[j+1]*pow_lhs_fft[j+1]);
      if(fabs(pow_lhs_fft[j+1]) < 1e-30 && fabs(pow_lhs_fft[j]) < 1e-30)
	temp2[i]= 0.0;
      else
	temp2[i]= (180./M_PI)*atan2(pow_lhs_fft[j], pow_lhs_fft[j+1]);
    }
    for(i=0; i< freq_hi; i++) {
      pow_lhs_fft[i]= temp1[i];
      pow_lhs_fft[freq_hi+i]= temp2[i];
    }
    
    /**** Computing mag and phase of the voltage signal ***/
    
    temp1[0]= fabs(phi_lhs_fft[0]);
    temp2[0]= (phi_lhs_fft[0] > 0.0) ? 0.0 : 180.0;
    for(i=1, j=2; i< freq_hi; i++, j+=2) {
      temp1[i]= 2*sqrt(phi_lhs_fft[j]*phi_lhs_fft[j]+ phi_lhs_fft[j+1]*phi_lhs_fft[j+1]);
      if(fabs(phi_lhs_fft[j+1]) < 1e-30 && fabs(phi_lhs_fft[j]) < 1e-30)
	temp2[i]= 0.0;
      else
	temp2[i]= (180./M_PI)*atan2(phi_lhs_fft[j], phi_lhs_fft[j+1]);
    }
    for(i=0; i< freq_hi; i++) {
      phi_lhs_fft[i]= temp1[i];
      phi_lhs_fft[freq_hi+i]= temp2[i];
    }
  } 

  /********** Then do the RHS stuff     ***********/
  if(rhs_flag) {
    for(i=0; i< nfft; i++) {
      cur_rhs_fft[i]= cur_rhs_hist[i];
      pow_rhs_fft[i]= pow_rhs_hist[i];
      phi_rhs_fft[i]= phi_rhs_hist[i];
    }
    realft(phi_rhs_fft-1, freq_hi, 1);
    realft(pow_rhs_fft-1, freq_hi, 1);
    realft(cur_rhs_fft-1, freq_hi, 1);
    
    /**** Computing mag and phase of the current signal ***/
    
    temp1[0]= fabs(cur_rhs_fft[0]);
    temp2[0]= (cur_rhs_fft[0] > 0.0) ? 0.0 : 180.0;
    for(i=1, j=2; i< freq_hi; i++, j+=2) {
      temp1[i]= 2*sqrt(cur_rhs_fft[j]*cur_rhs_fft[j]+ cur_rhs_fft[j+1]*cur_rhs_fft[j+1]);
      if(fabs(cur_rhs_fft[j+1]) < 1e-30 && fabs(cur_rhs_fft[j]) < 1e-30)
	temp2[i]= 0.0;
      else
	temp2[i]= (180./M_PI)*atan2(cur_rhs_fft[j], cur_rhs_fft[j+1]);
    }
    for(i=0; i< freq_hi; i++) {
      cur_rhs_fft[i]= temp1[i];
      cur_rhs_fft[freq_hi+i]= temp2[i];
    }
    
    /**** Computing mag and phase of the power signal *****/
    
    temp1[0]= fabs(pow_rhs_fft[0]);
    temp2[0]= (pow_rhs_fft[0] > 0.0) ? 0.0 : 180.0;
    for(i=1, j=2; i< freq_hi; i++, j+=2) {
      temp1[i]= 2*sqrt(pow_rhs_fft[j]*pow_rhs_fft[j]+ pow_rhs_fft[j+1]*pow_rhs_fft[j+1]);
      if(fabs(pow_rhs_fft[j+1]) < 1e-30 && fabs(pow_rhs_fft[j]) < 1e-30)
	temp2[i]= 0.0;
      else
	temp2[i]= (180./M_PI)*atan2(pow_rhs_fft[j], pow_rhs_fft[j+1]);
    }
    for(i=0; i< freq_hi; i++) {
      pow_rhs_fft[i]= temp1[i];
      pow_rhs_fft[freq_hi+i]= temp2[i];
    }
    
    /**** Computing mag and phase of the voltage signal ***/
    
    temp1[0]= fabs(phi_rhs_fft[0]);
    temp2[0]= (phi_rhs_fft[0] > 0.0) ? 0.0 : 180.0;
    for(i=1, j=2; i< freq_hi; i++, j+=2) {
      temp1[i]= 2*sqrt(phi_rhs_fft[j]*phi_rhs_fft[j]+ phi_rhs_fft[j+1]*phi_rhs_fft[j+1]);
      if(fabs(phi_rhs_fft[j+1]) < 1e-30 && fabs(phi_rhs_fft[j]) < 1e-30)
	temp2[i]= 0.0;
      else
	temp2[i]= (180./M_PI)*atan2(phi_rhs_fft[j], phi_rhs_fft[j+1]);
    }
    for(i=0; i< freq_hi; i++) {
      phi_rhs_fft[i]= temp1[i];
      phi_rhs_fft[freq_hi+i]= temp2[i];
    }
  }

  if(ybc_flag == PERIODIC) {
    /* djc new * only shows pos. freq part*/
    for(k=0;k<=khi;k++) {
      four1(phi_mid_k_t[k]-1, nfft, 1);
      for(i=0;i<freq_hi;i++)
	phi_mid_k_f[k][i] = sqrt( phi_mid_k_t[k][2*i]*phi_mid_k_t[k][2*i]
				 +phi_mid_k_t[k][2*i+1]*phi_mid_k_t[k][2*i+1]);
    }
  }
}

/***************************************************************/

